REM
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM    *  Grumpy, the most grouchy http server out there!  *
REM    *  Copyright (C) Mateusz Viste 2008                 *
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM
REM   ----------------------------------------------------------------------
REM    This program is free software: you can redistribute it and/or modify
REM    it under the terms of the GNU General Public License as published by
REM    the Free Software Foundation, either version 3 of the License, or
REM    (at your option) any later version.
REM
REM    This program is distributed in the hope that it will be useful,
REM    but WITHOUT ANY WARRANTY; without even the implied warranty of
REM    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
REM    GNU General Public License for more details.
REM
REM    You should have received a copy of the GNU General Public License
REM    along with this program.  If not, see <http://www.gnu.org/licenses/>.
REM   ----------------------------------------------------------------------
REM
REM  To-do:
REM   - better detection of 3rd parameter of a GET
REM   - some cache controls
REM   - HEAD implementation
REM   - implement some 304 responses (when "If-Modified-Since:")
REM   - add "if-unmodified-since:" support (412 code)
REM   - add the DATE header to responses
REM   - add partial download support (accept-ranges: bytes)
REM   - add "Last-Modified:" header whenever possible
REM   - Sort directory's entries before displaying
REM   - nicer directory listing layout
REM

#INCLUDE "file.bi"

DECLARE SUB ReturnFile(FileToReturn AS STRING)
DECLARE SUB Return301(RedirectTo AS STRING)
DECLARE SUB Return400()
DECLARE SUB Return403()
DECLARE SUB Return404(MissingFile AS STRING)
DECLARE SUB Return500()
DECLARE SUB Return501()
DECLARE SUB Return505()
DECLARE SUB ListDir(DirectoryToList AS STRING)
DECLARE FUNCTION RemoveDoubleSlash(StringToRemove AS STRING) AS STRING
DECLARE FUNCTION ReadCFG(CFGfile AS STRING, CFGField AS STRING) AS STRING
DECLARE FUNCTION CheckForEvasion(JailPath AS STRING, EvadingFile AS STRING) AS BYTE
DECLARE FUNCTION IsDirectory(ElementToCheck AS STRING) AS BYTE

CONST pVer AS STRING = "0.10"

' Initializing buffers:
DIM SHARED AS STRING LineBuff, OperationBuff, HostBuff, UserAgentBuff, AcceptBuff, AcceptLanguageBuff,_
              AcceptEncodingBuff, AcceptCharsetBuff, KeepAliveBuff, ConnectionBuff, CookieBuff,_
              RefererBuff
DIM SHARED AS STRING GetQuery, GetParam1, GetParam2
' Other variables:
DIM SHARED AS BYTE VerboseMode, AllowDirListing
DIM SHARED AS STRING Token, Value, RootPath, LocalFile

VerboseMode = VAL(ReadCFG("/etc/grumpy.cfg", "Verbose"))
AllowDirListing = VAL(ReadCFG("/etc/grumpy.cfg", "AllowDirListing"))
RootPath = ReadCFG("/etc/grumpy.cfg", "RootDir")
IF RootPath = "" THEN
  RootPath = "/var/www/"
  VerboseMode = 1
  AllowDirListing = 0
END IF
IF RIGHT(RootPath, 1) <> "/" THEN RootPath += "/"

OPEN CONS FOR INPUT AS #4
OPEN CONS FOR OUTPUT AS #5
IF VerboseMode > 0 THEN
  OPEN "/var/log/grumpy.log" FOR APPEND AS #10
  PRINT #10, "" : PRINT #10, "---[ "; Date ; " / "; time; " ]---"
END IF

DO
  LINE INPUT #4, LineBuff
  IF VerboseMode = 2 THEN PRINT #10, LineBuff
  Token = UCASE(LEFT(LineBuff, INSTR(LineBuff, " ") - 1))
  Value = TRIM(MID(LineBuff, INSTR(LineBuff, " ") + 1))
  SELECT CASE Token
    CASE "GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "TRACE", "OPTIONS"
      OperationBuff = LineBuff
    CASE "HOST:"
      HostBuff = Value
    CASE "USER-AGENT:"
      UserAgentBuff = Value
    CASE "ACCEPT:"
      AcceptBuff = Value
    CASE "ACCEPT-LANGUAGE:"
      AcceptLanguageBuff = Value
    CASE "ACCEPT-ENCODING:"
      AcceptEncodingBuff = Value
    CASE "ACCEPT-CHARSET:"
      AcceptCharsetBuff = Value
    CASE "KEEP-ALIVE:"
      KeepAliveBuff = Value
    CASE "CONNECTION:"
      ConnectionBuff = Value
    CASE "REFERER:"
      RefererBuff = Value
    CASE "COOKIE:"
      CookieBuff = Value
  END SELECT
LOOP UNTIL LEN(LineBuff) = 0

REM * Process the request...

GetQuery = MID(OperationBuff, 1, INSTR(OperationBuff, " ") - 1)
GetParam1 = MID(MID(OperationBuff, LEN(GetQuery) + 2), 1, INSTR(MID(OperationBuff, LEN(GetQuery) + 2), " ") - 1)
GetParam2 = MID(OperationBuff, LEN(GetQuery) + 3 + LEN(GetParam1))

IF VerboseMode > 0 THEN PRINT #10, "Query="; GetQuery; " Param1="; GetParam1; " Param2="; GetParam2

SELECT CASE GetQuery
  CASE "GET"
    SELECT CASE GetParam2
      CASE "HTTP/1.0", "HTTP/1.1"
        LocalFile = RemoveDoubleSlash(RootPath + "/" + GetParam1)
        IF CheckForEvasion(RootPath, LocalFile) = 1 THEN
            Return403()
          ELSE
            IF IsDirectory(LocalFile) = 1 THEN
              IF RIGHT(LocalFile, 1) = "/" THEN
                    IF FileExists(LocalFile + "index.htm") <> 0 THEN
                        LocalFile += "index.htm"
                        ReturnFile(LocalFile)
                      ELSEIF FileExists(LocalFile + "index.html") <> 0 THEN
                        LocalFile += "index.html"
                        ReturnFile(LocalFile)
                      ELSEIF AllowDirListing = 1 THEN
                        ListDir(LocalFile)
                      ELSE
                        Return403()
                    END IF
                  ELSE
                    Return301(GetParam1 + "/")
                END IF
              ELSE
                IF FileExists(LocalFile) <> 0 THEN
                    ReturnFile(LocalFile)
                  ELSE
                    Return404(GetParam1)
                END IF
          END IF
        END IF
      CASE ELSE
        Return505()
    END SELECT
  CASE "HEAD", "POST", "PUT", "DELETE", "CONNECT", "TRACE", "OPTIONS"
    Return501()
  CASE ELSE
    Return400()
END SELECT

CLOSE #4
CLOSE #5
IF VerboseMode > 0 THEN CLOSE #10


REM ***  END OF THE MAIN PROGRAM  ***


SUB Return404(MissingFile AS STRING) ' "not found"
  IF VerboseMode > 0 THEN PRINT #10, "Response: 404 ("; MissingFile; ")"
  PRINT #5, "HTTP/1.1 404 "; MissingFile
  'PRINT #5, "Date: Mon, 29 Sep 2008 06:48:35 GMT"
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 1030"
  PRINT #5, "Content-Type: text/html;charset=utf-8"
  PRINT #5,
  PRINT #5, "<html><head><title>404</title></head><body><h1>404 - Not found</h1></body></html>"
END SUB


SUB ReturnFile(FileToReturn AS STRING) ' "ok"
  DIM AS INTEGER x, FileLength
  DIM AS STRING*1 ByteBuff
  IF VerboseMode > 0 THEN PRINT #10, "Response: 200 ("; FileToReturn; ")"

  OPEN FileToReturn FOR BINARY AS #20
  PRINT #5, "HTTP/1.1 200 OK"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  PRINT #5, "Content-Length: " & LOF(20)
  'PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,

  FOR x = 1 TO LOF(20)
    GET #20, x, ByteBuff
    PRINT #5, ByteBuff;
  NEXT x
  CLOSE #20
END SUB


SUB Return403() ' "forbidden"
  IF VerboseMode > 0 THEN PRINT #10, "Response: 403"
  PRINT #5, "HTTP/1.1 403"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  PRINT #5, "<html><head><title>403</title></head><body><h1>403 - Forbidden</h1></body></html>"
END SUB


SUB Return501() ' "not implemented"
  IF VerboseMode > 0 THEN PRINT #10, "Response: 501"
  PRINT #5, "HTTP/1.1 501"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  PRINT #5, "<html><head><title>501</title></head><body><h1>501 - Not implemented</h1></body></html>"
END SUB


SUB Return400() ' "bad request"
  IF VerboseMode > 0 THEN PRINT #10, "Response: 400"
  PRINT #5, "HTTP/1.1 400"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  PRINT #5, "<html><head><title>400</title></head><body><h1>400 - Bad request</h1></body></html>"
END SUB


SUB Return505() ' "HTTP version not supported"
  IF VerboseMode > 0 THEN PRINT #10, "Response: 505"
  PRINT #5, "HTTP/1.1 505"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  PRINT #5, "<html><head><title>505</title></head><body><h1>505 - HTTP version not supported</h1></body></html>"
END SUB


SUB Return500() ' "Internal server error"
  IF VerboseMode > 0 THEN PRINT #10, "Response: 500"
  PRINT #5, "HTTP/1.1 500"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  PRINT #5, "<html><head><title>500</title></head><body><h1>500 - Internal server error</h1></body></html>"
END SUB


SUB ListDir(DirectoryToList AS STRING)
  DIM AS STRING FileName, ListTitle
  DIM AS UINTEGER FileAttrib
  IF VerboseMode > 0 THEN PRINT #10, "Response: 200 (list directory """; DirectoryToList; """)"
  ListTitle = MID(DirectoryToList, LEN(RootPath))
  PRINT #5, "HTTP/1.1 200"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  PRINT #5, "<!DOCTYPE HTML PUBLIC ""-//W3C//DTD HTML 4.01 Transitional//EN"" ""http://www.w3.org/TR/html4/loose.dtd"">"
  PRINT #5, "<html><head><title>"; ListTitle; "</title></head><body bgcolor=""#D0D0FF""><br><table cellpadding=""10"" width=""100%"" border=""1""><tr><td><b><big>Directory: "; ListTitle; "</big></b><hr><br><table>"

  FileName = DIR(DirectoryToList + "*", &h37, FileAttrib)
  DO
    IF FileName <> "." AND (FileName <> ".." OR DirectoryToList <> RootPath) THEN
      IF FileAttrib AND &h10 THEN
          PRINT #5, "<tr><td><a href="""; FileName; "/"">"; FileName; "</a></td><td>&nbsp;&lt;DIR&gt;</tr>"
        ELSE
          PRINT #5, "<tr><td><a href="""; FileName; """ target=""_blank"">"; FileName; "</a></td><td></td></tr>"
      END IF
    END IF
    FileName = DIR(FileAttrib)
  LOOP UNTIL FileName = ""

  PRINT #5, "</table></td></tr></table>"
  PRINT #5, "<i><small>Grumpy HTTP server v"; pVer; "</small></i>"
  PRINT #5, "</body></html>"
END SUB


FUNCTION RemoveDoubleSlash(StringToCheck AS STRING) AS STRING
  DIM AS STRING Wynik
  DIM AS INTEGER x
  Wynik = StringToCheck
  DO
    x = INSTR(Wynik, "//")
    IF x > 0 THEN Wynik = MID(Wynik, 1, x) + MID(Wynik, x + 2)
  LOOP UNTIL INSTR(Wynik, "//") = 0
  RETURN Wynik
END FUNCTION


FUNCTION ReadCFG(CFGfile AS STRING, CFGField AS STRING) AS STRING
 STATIC CfgTable(1 TO 2, 0 TO 255) AS STRING
 DIM AS STRING CfgReturnString, CfgTmpBuffer
 DIM AS INTEGER Counter

 IF CfgTable(1, 0) <> "init ok" THEN
   DIM AS INTEGER CfgFileHandler, CfgColonPos, Counter
   CfgFileHandler = FREEFILE
   CfgReturnString = ""
   Counter = 0
   CfgTable(1, 0) = "init ok"
   IF DIR(CFGfile) <> "" THEN
      OPEN CFGfile FOR INPUT AS #CfgFileHandler
      DO
         Counter += 1
         LINE INPUT #CfgFileHandler, CfgTmpBuffer
         IF MID(TRIM(CfgTmpBuffer), 1, 1) <> "#" THEN
           CfgColonPos = INSTR(CfgTmpBuffer, "=")
           CfgTable(1, Counter) = TRIM(MID(CfgTmpBuffer, 1, CfgColonPos - 1))
           CfgTable(2, Counter) = TRIM(MID(CfgTmpBuffer, CfgColonPos + 1))
         END IF
      LOOP UNTIL EOF(CfgFileHandler) OR Counter = 255
      CLOSE #CfgFileHandler
   END IF
   CfgTable(2, 0) = STR(Counter)
 END IF

 Counter = 0
 DO
   Counter += 1
   IF UCASE(CfgTable(1, Counter)) = UCASE(CFGField) THEN CfgReturnString = CfgTable(2, Counter)
 LOOP UNTIL CfgReturnString <> "" OR Counter >= VAL(CfgTable(2, 0))

 RETURN CfgReturnString
END FUNCTION


FUNCTION CheckForEvasion(JailPath AS STRING, EvadingFile AS STRING) AS BYTE
  DIM AS BYTE Wynik
  IF INSTR(EvadingFile, "..") > 0 THEN Wynik = 1 ELSE Wynik = 0
  RETURN Wynik
END FUNCTION


FUNCTION IsDirectory(ElementToCheck AS STRING) AS BYTE
  DIM AS BYTE Wynik
  DIM AS UINTEGER GetAttribs
  DIM AS STRING TempString
  TempString = DIR(ElementToCheck, &h37, GetAttribs)
  IF GetAttribs AND &h10 THEN Wynik = 1 ELSE Wynik = 0
  Return Wynik
END FUNCTION


SUB Return301(RedirectTo AS STRING)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 301 ("; RedirectTo; ")"
  PRINT #5, "HTTP/1.1 301"
  'PRINT #5, "Date: Mon, 29 Sep 2008 09:45:56 GMT
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Location: "; RedirectTo
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  PRINT #5, "<html><head><title>301</title></head><body><h1>301 - Moved Permanently</h1><br>The document has moved <a href="""; RedirectTo; """>here</a>.</body></html>"
END SUB
